from datetime import datetime
from pathlib import Path
import json
import shutil
from typing import Any, Dict, Optional, Sequence, Union
import warnings

from .data_processors import DEFAULT_DATA_PROCESSORS, DataProcessor
from .data_folder_tools import (
    DEFAULT_FOLDER_PATTERN,
    create_data_folder,
    generate_data_folder_relative_pathname,
    get_latest_data_folder,
)


__all__ = ["save_data", "DataHandler"]

NODE_FILENAME = "node.json"


def save_data(
    data_folder: Path,
    data: Dict[str, Any],
    node_contents: Dict[str, Any],
    metadata: Optional[Dict[str, Any]] = None,
    data_filename: str = "data.json",
    data_processors: Sequence[DataProcessor] = (),
) -> Path:
    """Save data to a folder

    The data (assumed to be a dict) is saved as a json file to "{data_folder}/{data_filename}", which typically
    follows the format "%Y-%m-%d/#{idx}_{name}_%H%M%S/data.json".
    Non-serialisable contents in data such as figures and arrays are saved into separate files and the paths are
    referenced from the data dictionary.

    This function also applies a list of data processors to the data before saving it. The data processors are
    applied in the order they are provided.

    This function is used by the DataHandler class to save data to a folder.

    :param data_folder: The folder where the data will be saved
    :param data: The data to be saved
    :param node_contents: The contents of the node file
    :param metadata: Metadata to be saved
    :param data_filename: The filename of the data
    :param data_processors: A list of data processors to be applied to the data
    :return: The path of the saved data folder
    """
    if isinstance(data_folder, str):
        data_folder = Path(data_folder)

    if not data_folder.exists():
        raise NotADirectoryError(f"Save_data: data_folder {data_folder} does not exist")

    if (data_folder / NODE_FILENAME).exists():
        raise FileExistsError(f"Save_data: data_folder {data_folder} already contains data")

    node_contents = node_contents.copy()
    if metadata is not None:
        node_contents.setdefault("metadata", {})
        node_contents["metadata"].update(metadata)
    json_node_contents = json.dumps(node_contents, indent=4)
    (data_folder / NODE_FILENAME).write_text(json_node_contents)

    if not isinstance(data, dict):
        raise TypeError("save_data: 'data' must be a dictionary")

    processed_data = data.copy()
    for data_processor in data_processors:
        processed_data = data_processor.process(processed_data)

    json_data = json.dumps(processed_data, indent=4)
    (data_folder / data_filename).write_text(json_data)

    for data_processor in data_processors:
        data_processor.post_process(data_folder=data_folder)

    return data_folder


class DataHandler:
    """A class to handle data saving.

    This class provides functionality to save data to a specified data folder.
    It allows for the creation of a new data folder, saving data to the folder,
    and applying data processors to the saved data.

    :param name: The name of the data handler.
    :type name: str, optional
    :param data_processors: The data processors to be applied to the saved data.
    :type data_processors: Sequence[DataProcessor], optional
    :param root_data_folder: The root folder where the data will be saved.
    :type root_data_folder: str or Path, optional
    :param folder_pattern: The pattern used to create the data folder.
    :type folder_pattern: str, optional
    :param path: The path to the data folder.
    :type path: Path, optional

    Example usage:

    .. code-block:: python

        data_handler = DataHandler("T1_experiment")
        data = {"T1": 1e-6, "T1_arr": np.array([1, 2, 3]), "T1_fig": plt.figure()}
        data_handler.save_data(data)
    """

    default_data_processors = DEFAULT_DATA_PROCESSORS
    root_data_folder: Path = None
    folder_pattern: str = DEFAULT_FOLDER_PATTERN
    data_filename: str = "data.json"
    additional_files: Dict[str, str] = {}
    node_data: Dict[str, Any] = {}

    def __init__(
        self,
        name: Optional[str] = None,
        data_processors: Optional[Sequence[DataProcessor]] = None,
        root_data_folder: Optional[Union[str, Path]] = None,
        folder_pattern: Optional[str] = None,
        additional_files: Optional[Dict[str, str]] = None,
    ):
        self.name = name
        if data_processors is not None:
            self.data_processors = data_processors
        else:
            self.data_processors = [processor() for processor in self.default_data_processors]

        if root_data_folder is not None:
            self.root_data_folder = root_data_folder
        if folder_pattern is not None:
            self.folder_pattern = folder_pattern
        if additional_files is not None:
            self.additional_files = additional_files

        self.path = None
        self.path_properties = None

    def generate_node_contents(
        self,
        idx: Optional[int] = None,
        created_at: Optional[datetime] = None,
        metadata: Optional[Dict[str, Any]] = None,
    ) -> Dict[str, Any]:
        """Generate the contents of the node.json file, which contains all node-related metadata.

        :param idx: The index of the data folder.
        :type idx: int, optional
        :param created_at: The datetime to be used in the folder name.
        :type created_at: datetime, optional
        :param metadata: The user-specified metadata associated with the data. "name" and "data_path" are added to
        this metadata.
        :type metadata: any, optional
        :return: The contents of the node.json file.
        :rtype: dict
        """
        # Check if an empty folder has been created, if so, use the idx and datetime from the folder
        if self.path_properties is not None and idx is None and created_at is None:
            if not (self.path / NODE_FILENAME).exists():
                idx = self.path_properties["idx"]
                created_at = self.path_properties["created_at"]

        if idx is None:
            latest_folder_properties = get_latest_data_folder(self.root_data_folder, folder_pattern=self.folder_pattern)
            idx = latest_folder_properties["idx"] + 1 if latest_folder_properties is not None else 1
        if created_at is None:
            created_at = datetime.now()

        metadata = metadata.copy() if metadata is not None else {}
        metadata["name"] = self.name

        metadata["data_path"] = generate_data_folder_relative_pathname(
            idx=idx, name=self.name, created_at=created_at, folder_pattern=self.folder_pattern
        )

        return {
            "created_at": created_at.astimezone().isoformat(timespec="seconds"),
            "metadata": metadata,
            "data": self.node_data,  # TODO Add self.node_data
            "id": idx,
            "parents": [idx - 1] if idx > 1 else [],
        }

    def create_data_folder(
        self,
        name: Optional[str] = None,
        idx: Optional[int] = None,
        created_at: Optional[datetime] = None,
        create: bool = True,
    ) -> Dict[str, Union[str, int]]:
        """Create a new data folder in the root data folder.

        This method creates a new data folder in the root data folder specified
        in the `root_data_folder` attribute. The name of the data folder can be
        specified using the `name` parameter. An index can also be provided using
        the `idx` parameter. If a datetime object is provided using the `created_at`
        parameter, it will be used in the folder name. By default, the data folder
        is created.

        :param name: The name of the data folder.
        :type name: str, optional
        :param idx: The index of the data folder.
        :type idx: int, optional
        :param created_at: The datetime to be used in the folder name.
        :type created_at: datetime, optional
        :param create: Whether to create the data folder or not.
        :type create: bool, optional
        :return: The properties of the created data folder.
        :rtype: dict
        :raises ValueError: If the name is not specified.

        Example usage:

        .. code-block:: python

            data_handler.create_data_folder(name="T1_experiment", idx=1, created_at=datetime.now())
        """
        if name is not None:
            self.name = name
        if self.name is None:
            raise ValueError("DataHandler: name must be specified")
        if self.root_data_folder is None:
            raise ValueError("DataHandler: root_data_folder must be specified")

        self.path_properties = create_data_folder(
            root_data_folder=self.root_data_folder,
            folder_pattern=self.folder_pattern,
            created_at=created_at,
            name=self.name,
            idx=idx,
            create=create,
        )
        self.path = self.path_properties["path"]
        return self.path_properties

    def save_data(
        self,
        data,
        name: Optional[str] = None,
        metadata: Optional[Dict[str, Any]] = None,
        idx: Optional[int] = None,
        node_contents: Optional[Dict[str, Any]] = None,
        created_at: Optional[datetime] = None,
    ):
        """Save the data to the data folder.

        This method saves the provided data to the data folder specified in the
        `path` attribute. The name of the data folder can be specified using the
        `name` parameter. The metadata associated with the data can be provided
        using the `metadata` parameter. An index can also be provided using the
        `idx` parameter. If a datetime object is provided using the `created_at`
        parameter, it will be used in the folder name.

        A new data folder is created if
        - the `path` attribute is not set
        - the `path` attribute is set and the data folder already contains data

        :param data: The data to be saved.
        :type data: any
        :param name: The name of the data folder.
        :type name: str, optional
        :param metadata: The metadata associated with the data.
        :type metadata: any, optional
        :param idx: The index of the data folder.
        :type idx: int, optional
        :param created_at: The datetime to be used in the folder name.
        :type created_at: datetime, optional
        :raises ValueError: If the name is not specified.
        :return: The result of saving the data.
        :rtype: any

        Example usage:

        .. code-block:: python

            data_handler.save_data(data, name="T1_experiment", metadata=metadata, idx=1, created_at=datetime.now())
        """
        if name is not None:
            self.name = name
        if self.name is None:
            raise ValueError("DataHandler: name must be specified")

        if node_contents is None:
            # Generate for a new data folder or use existing node_contents if self.path is set and folder is empty
            node_contents = self.generate_node_contents(idx=idx, created_at=created_at, metadata=metadata)
        elif created_at is not None:
            warnings.warn("DataHandler: created_at is ignored when node_contents is provided", UserWarning)
        elif idx is not None and idx != node_contents["id"]:
            warnings.warn("DataHandler: idx is ignored when node_contents is provided", UserWarning)

        idx = node_contents["id"]
        created_at = datetime.fromisoformat(node_contents["created_at"])

        # At this point we have determined the idx and datetime to use for the data folder
        # It may be that a folder already exists with the same idx and datetime and is populated.
        # In this case we should raise an error

        # Verify that the data folder has not been created and populated

        data_path = Path(self.root_data_folder) / node_contents["metadata"]["data_path"]
        if data_path.exists():
            if (data_path / NODE_FILENAME).exists():
                raise FileExistsError(f"DataHandler: data folder {data_path} already contains data")
            else:
                # The folder has been created but not populated, so we can use it
                pass
        else:
            self.create_data_folder(name=self.name, idx=idx, created_at=created_at)

        data_folder = save_data(
            data_folder=self.path,
            data=data,
            data_filename=self.data_filename,
            node_contents=node_contents,
            data_processors=self.data_processors,
        )

        for source_name, destination_name in self.additional_files.items():
            if not Path(source_name).exists():
                warnings.warn(
                    f"Additional file {source_name} does not exist, not copying",
                    UserWarning,
                )
                continue

            shutil.copy(source_name, data_folder / destination_name)

        return data_folder



"""
CALIBRATIONS
"""

from qm.qua import *
from qm import QuantumMachinesManager
from qm import QuantumMachine
from qualang_tools.plot import interrupt_on_close
from qualang_tools.results import progress_counter, fetching_tool
from qualang_tools.units import unit
from qualang_tools.loops import from_array
from qm import SimulationConfig
import matplotlib.pyplot as plt
import numpy as np
from scipy import signal


available_variables = ["frequency", "amplitude", "duration"]
u = unit()


class QUA_calibrations:
    """
    Allows to easily perform various single qubit calibrations with the OPX.
    """

    def _check_element(self, element):
        if element in self.config["elements"]:
            return True
        else:
            raise Exception(f"The element '{element}' is not in the current config.")

    def _check_operation(self, operation, element):
        if operation in self.config["elements"][element]["operations"]:
            return True
        else:
            raise Exception(f"The operation '{operation}' is not in the current config for element '{element}'.")

    def _check_iw(self, int_weights):
        for iw in int_weights:
            if (
                iw
                not in self.config["pulses"][self.config["elements"][self.readout_elmt]["operations"][self.readout_op]][
                    "integration_weights"
                ]
            ):
                raise Exception(
                    f"The integration weight '{iw}' is not in the current config for operation '{self.readout_op}'."
                )
        return True

    def _check_outputs(self, outputs):
        for out in outputs:
            if out not in self.config["elements"][self.readout_elmt]["outputs"]:
                raise Exception(f"The output '{out}' is not in the current config for element '{self.readout_elmt}'.")
        return True

    def _check_plot_options(self, plot_options):
        if type(plot_options) is not dict:
            raise Exception("plot_options must be a python dictionary.")
        for key in plot_options:
            if key in self.plot_options.keys():
                self.plot_options[key] = plot_options[key]
            else:
                raise Exception(
                    f"Plot option {key} is not implemented, available options are {self.plot_options.keys()}"
                )

    def _user_variables_checks(self, calibration, scan_variables, iterations):
        # Check calibration
        if calibration in self.calibration_list:
            raise Exception("For the moment, it is not possible to run the same calibration twice in a row.")
        # Check variables
        for var in scan_variables:
            if var[0] not in available_variables:
                raise Exception(
                    f"The {var[0]} scan is not implemented. The available scan variables are {available_variables}."
                )
        # Check iterations
        if not (type(iterations) is int and iterations > 0):
            raise Exception("interations must be a python positive integer.")

    def _user_variables_init(self, calibration, scan_variables, iterations, cooldown_time):
        self.calibration_list.append(calibration)
        self.results[calibration] = {
            "scan_variables": [],
            "amplitude": [],
            "phase": [],
            "fit": {},
        }
        self.scan_var[calibration] = {}
        self.user_specifics[calibration] = {}
        # Add variables
        for var in scan_variables:
            self.scan_var[calibration][var[0]] = var[1]
            self.results[calibration]["scan_variables"].append(var)
        # Add iterations
        self.scan_var[calibration]["averaging"] = iterations
        # Check cooldown time
        if cooldown_time is None or (cooldown_time >= 4 and type(cooldown_time) is int):
            self.user_specifics[calibration]["cooldown_time"] = cooldown_time
        elif cooldown_time == 0:
            self.user_specifics[calibration]["cooldown_time"] = None
        else:
            raise Exception("cooldown_time must be a python positive integer larger than 4.")

    def _get_qua_variables(self, current_calib):
        n = declare(int)  # Averaging index
        var1 = None
        var1_scan = None
        var1_name = None
        var2 = None
        var2_scan = None
        var2_name = None
        for var_name in self.scan_var[current_calib].keys():
            if var_name == "frequency":
                if var1 is None:
                    var1 = declare(int)
                    var1_scan = self.scan_var[current_calib][var_name]
                    var1_name = var_name
                else:
                    var2 = declare(int)
                    var2_scan = self.scan_var[current_calib][var_name]
                    var2_name = var_name
            elif var_name == "amplitude":
                if current_calib == "Ramsey":
                    raise Exception("Amplitude scan is not available for Ramsey measurements.")
                elif (current_calib == "Resonator_spec") and self.flux_line_elmt is None:
                    raise Exception("Amplitude scan is not available for resonator spectroscopy without flux line.")
                if var1 is None:
                    var1 = declare(fixed)
                    var1_scan = self.scan_var[current_calib][var_name]
                    var1_name = var_name
                else:
                    var2 = declare(fixed)
                    var2_scan = self.scan_var[current_calib][var_name]
                    var2_name = var_name
            elif var_name == "duration":
                if (self.scan_var[current_calib][var_name] < 4).any():
                    raise Exception("No scanned duration can be lower than 4 clock cycles.")
                if current_calib in ["Rabi", "Ramsey"]:
                    pulse = self.config["elements"][self.qubit_elmt]["operations"][self.qubit_op]
                    for inpt in self.config["pulses"][pulse]["waveforms"]:
                        wf = self.config["pulses"][pulse]["waveforms"][inpt]
                        if self.config["waveforms"][wf]["type"] == "arbitrary":
                            if (
                                self.scan_var[current_calib][var_name] < self.config["pulses"][pulse]["length"] // 4
                            ).any():
                                raise Exception(
                                    "The scanned duration must be greater than the one defined in the configuration for arbitrary waveforms. Arbitrary waveforms can only be stretched and not compressed on the fly."
                                )
                if var1 is None:
                    var1 = declare(int)
                    var1_scan = self.scan_var[current_calib][var_name]
                    var1_name = var_name
                else:
                    var2 = declare(int)
                    var2_scan = self.scan_var[current_calib][var_name]
                    var2_name = var_name

            if len(self.scan_var[current_calib].keys()) == 2:
                return n, var1, var1_scan, var1_name
        return n, var1, var1_scan, var1_name, var2, var2_scan, var2_name

    def _align_measure_save(self, I, Q, I_st, Q_st, current_calib, align_flag=True):
        if align_flag:
            align(self.qubit_elmt, self.readout_elmt)
        # Measure the resonator
        if len(self.out) == 2:
            measure(
                self.readout_op,
                self.readout_elmt,
                None,
                dual_demod.full(self.iw[0], self.out[0], self.iw[1], self.out[1], I),
                dual_demod.full(self.iw[2], self.out[0], self.iw[3], self.out[1], Q),
            )
        elif len(self.out) == 1:
            measure(
                self.readout_op,
                self.readout_elmt,
                None,
                demod.full(self.iw[0], I, self.out[0]),
                demod.full(self.iw[1], Q, self.out[0]),
            )
        # Wait for the resonator to cooldown
        if self.user_specifics[current_calib]["cooldown_time"] is not None:
            wait(
                int(self.user_specifics[current_calib]["cooldown_time"]),
                self.readout_elmt,
            )
        # Save data to the stream processing
        save(I, I_st)
        save(Q, Q_st)

    def _plot_2d(self, x, y, amplitude, phase, x_label, y_label, title):
        plt.suptitle(title, fontsize=self.plot_options["fontsize"] + 2)
        plt.subplot(211)
        plt.cla()
        plt.title("Amplitude [V]")
        plt.pcolor(x, y, amplitude)
        plt.ylabel(y_label, fontsize=self.plot_options["fontsize"])
        plt.xticks(fontsize=self.plot_options["fontsize"])
        plt.yticks(fontsize=self.plot_options["fontsize"])
        plt.subplot(212)
        plt.cla()
        plt.title("Phase [rad]")
        plt.pcolor(x, y, phase)
        plt.xlabel(x_label, fontsize=self.plot_options["fontsize"])
        plt.ylabel(y_label, fontsize=self.plot_options["fontsize"])
        plt.xticks(fontsize=self.plot_options["fontsize"])
        plt.yticks(fontsize=self.plot_options["fontsize"])

    def _plot_1d(self, x, amplitude, phase, x_label, title):
        plt.suptitle(title, fontsize=self.plot_options["fontsize"] + 2)
        plt.subplot(211)
        plt.cla()
        plt.plot(
            x,
            amplitude,
            color=self.plot_options["color"],
            marker=self.plot_options["marker"],
            linewidth=self.plot_options["linewidth"],
        )
        plt.ylabel("Amplitude [V]", fontsize=self.plot_options["fontsize"])
        plt.xticks(fontsize=self.plot_options["fontsize"])
        plt.yticks(fontsize=self.plot_options["fontsize"])
        plt.subplot(212)
        plt.cla()
        plt.plot(
            x,
            phase,
            color=self.plot_options["color"],
            marker=self.plot_options["marker"],
            linewidth=self.plot_options["linewidth"],
        )
        plt.xlabel(x_label, fontsize=self.plot_options["fontsize"])
        plt.ylabel("Phase [rad]", fontsize=self.plot_options["fontsize"])
        plt.xticks(fontsize=self.plot_options["fontsize"])
        plt.yticks(fontsize=self.plot_options["fontsize"])

    def _get_IQ_data_and_plot(self, job, calib, plot, dim):
        pulse = self.config["elements"][self.readout_elmt]["operations"][self.readout_op]
        res_handles = job.result_handles
        if dim == 1:
            if plot == "live":
                results = fetching_tool(job, data_list=["I", "Q", "iteration"], mode="live")
                # Live plotting
                fig = plt.figure(figsize=self.plot_options["figsize"])
                interrupt_on_close(fig, job)  # Interrupts the job when closing the figure
                while results.is_processing():
                    I, Q, iteration = results.fetch_all()
                    I = u.demod2volts(I, self.config["pulses"][pulse]["length"])
                    Q = u.demod2volts(Q, self.config["pulses"][pulse]["length"])

                    progress_counter(
                        iteration,
                        self.scan_var[calib]["averaging"],
                        start_time=results.get_start_time(),
                    )

                    self.results[calib]["amplitude"] = np.sqrt(I**2 + Q**2)
                    self.results[calib]["phase"] = signal.detrend(np.unwrap(np.angle(I + 1j * Q)))

                    self._plot_1d(
                        x=self.scan_var[calib][list(self.scan_var[calib].keys())[0]],
                        amplitude=self.results[calib]["amplitude"],
                        phase=self.results[calib]["phase"],
                        x_label=list(self.scan_var[calib].keys())[0],
                        title=calib,
                    )
                    plt.pause(0.01)

            else:
                res_handles.wait_for_all_values()
                I = u.demod2volts(
                    res_handles.get("I").fetch_all(),
                    self.config["pulses"][pulse]["length"],
                )
                Q = u.demod2volts(
                    res_handles.get("Q").fetch_all(),
                    self.config["pulses"][pulse]["length"],
                )

                self.results[calib]["amplitude"] = np.sqrt(I**2 + Q**2)
                self.results[calib]["phase"] = signal.detrend(np.unwrap(np.angle(I + 1j * Q)))

                if plot == "full":
                    plt.figure(figsize=self.plot_options["figsize"])
                    self._plot_1d(
                        x=self.scan_var[calib][list(self.scan_var[calib].keys())[0]],
                        amplitude=self.results[calib]["amplitude"],
                        phase=self.results[calib]["phase"],
                        x_label=list(self.scan_var[calib].keys())[0],
                        title=calib,
                    )
        elif dim == 2:
            res_handles = job.result_handles
            if plot == "live":
                results = fetching_tool(job, data_list=["I", "Q", "iteration"], mode="live")

                # Live plotting
                fig = plt.figure(figsize=self.plot_options["figsize"])
                interrupt_on_close(fig, job)  # Interrupts the job when closing the figure
                while results.is_processing():
                    I, Q, iteration = results.fetch_all()
                    I = u.demod2volts(I, self.config["pulses"][pulse]["length"])
                    Q = u.demod2volts(Q, self.config["pulses"][pulse]["length"])

                    progress_counter(
                        iteration,
                        self.scan_var[calib]["averaging"],
                        start_time=results.get_start_time(),
                    )

                    self.results[calib]["amplitude"] = np.sqrt(I**2 + Q**2)
                    self.results[calib]["phase"] = signal.detrend(np.unwrap(np.angle(I + 1j * Q)))

                    self._plot_2d(
                        x=self.scan_var[calib][list(self.scan_var[calib].keys())[0]],
                        y=self.scan_var[calib][list(self.scan_var[calib].keys())[1]],
                        amplitude=self.results[calib]["amplitude"],
                        phase=self.results[calib]["phase"],
                        x_label=list(self.scan_var[calib].keys())[0],
                        y_label=list(self.scan_var[calib].keys())[1],
                        title=calib,
                    )
                    plt.pause(0.01)
            else:
                res_handles.wait_for_all_values()
                I = u.demod2volts(
                    res_handles.get("I").fetch_all(),
                    self.config["pulses"][pulse]["length"],
                )
                Q = u.demod2volts(
                    res_handles.get("Q").fetch_all(),
                    self.config["pulses"][pulse]["length"],
                )

                self.results[calib]["amplitude"] = np.sqrt(I**2 + Q**2)
                self.results[calib]["phase"] = signal.detrend(np.unwrap(np.angle(I + 1j * Q)))

                if plot == "full":
                    plt.figure(figsize=self.plot_options["figsize"])
                    self._plot_2d(
                        x=self.scan_var[calib][list(self.scan_var[calib].keys())[0]],
                        y=self.scan_var[calib][list(self.scan_var[calib].keys())[1]],
                        amplitude=self.results[calib]["amplitude"],
                        phase=self.results[calib]["phase"],
                        x_label=list(self.scan_var[calib].keys())[0],
                        y_label=list(self.scan_var[calib].keys())[1],
                        title=calib,
                    )

    def __init__(
        self,
        configuration,
        readout=("resonator", "readout"),
        qubit=("qubit", "pi"),
        integration_weights=("cos", "sin", "minus_sin", "cos"),
        outputs=("out1", "out2"),
        flux_line=None,
    ):
        """The QUA calibrations class. It has an API for performing various basic resonator and qubit calibrations.
        Currently the available calibrations are: 'resonator_spectroscopy', 'rabi' and 'ramsey'.
        All these calibrations are implemented with 1D or 2D scans among three variables ('frequency', 'amplitude', 'duration').
        The meaning of these variables is calibration specific.
        For instance 'duration' represent the qubit pulse length for Rabi and the idle time for Ramsey.
        Each calibration can be set using the `set_` (`set_rabi` for instance) method and executed using the
        `run_calibrations()` method or simulated using the `simulate_calibrations` method.
        The results are stored in a python dictionary called `results`.

        :param configuration: configuration file containing the qubit and readout elements and operations specified.
        :param readout: readout element and operation as defined in the configuration. Must be written as a tuple (element, operation).
        :param qubit: qubit element and operation as defined in the configuration. Must be written as a tuple (element, operation).
        :param integration_weights: integration weights as defined in the configuration. Must be written as a tuple of 4 items ('cos', 'sin', 'minus_sin', 'cos').
        :param outputs: readout element outputs as defined in the configuration. Must be a python tuple of two items ('out1', 'out2').
        :param flux_line: (optional) flux line element and operation as defined in the configuration. Must be written as a tuple (element, operation).
        """
        self.config = configuration
        # Check the validity of the user inputs for elements, operations, int weights and outputs
        if self._check_element(readout[0]):
            self.readout_elmt = readout[0]
        if self._check_element(qubit[0]):
            self.qubit_elmt = qubit[0]
        if flux_line is not None:
            if self._check_element(flux_line[0]):
                self.flux_line_elmt = flux_line[0]
            if self._check_operation(flux_line[1], flux_line[0]):
                self.flux_line_op = flux_line[1]
        else:
            self.flux_line_elmt = None
        if self._check_operation(readout[1], readout[0]):
            self.readout_op = readout[1]
        if self._check_operation(qubit[1], qubit[0]):
            self.qubit_op = qubit[1]
        if self._check_iw(integration_weights):
            self.iw = integration_weights
        if self._check_outputs(outputs):
            self.out = outputs

        self.calibration_list = []
        self.scan_var = {}
        self.results = {}
        self.cooldown_time = None
        self.user_specifics = {}
        self.simulation = False
        self.qm = None
        self.plot_options = {
            "fontsize": 14,
            "color": "b",
            "marker": None,
            "linewidth": 2,
            "figsize": (12, 15),
        }

    def set_resonator_spectroscopy(self, scan_variables, iterations=1, cooldown_time=None):
        """Sets the resonator spectroscopy calibration.
        This program will measure the resonator response for different readout pulse frequencies defined in scan_variables.

        :param scan_variables: variables to be scanned. Must be an array of tuples [('variable', [values])]. 1D and 2D scans are available.
        :param iterations: number of averaging iterations. Must be a positive Python integer.
        :param cooldown_time: resonator or qubit cooldown time in clock cycles (4ns). Must be an integer.
        :return: None
        """
        calibration = "Resonator_spec"
        self._user_variables_checks(calibration, scan_variables, iterations)
        self._user_variables_init(calibration, scan_variables, iterations, cooldown_time)

    def set_rabi(self, scan_variables, iterations=1, cooldown_time=None):
        """Performs a Rabi-like measurement by sending a pulse to the qubit and measuring the resonator response.
        The available scanning parameters are:
        'duration' (pulse length in clock cycles (4ns), must be an integer.),
        'frequency' (pulse frequency in Hz, must be an integer) and
        'amplitude' (pulse amplitude as a pre-factor to the amplitude set in the configuration, must be in [-2,2)).

        :param scan_variables: variables to be scanned. Must be an array of tuples [('variable', [values])]. 1D and 2D scans are available.
        :param iterations: number of averaging iterations. Must be a positive Python integer.
        :param cooldown_time: resonator or qubit cooldown time in clock cycles (4ns). Must be an integer.
        :return: None
        """
        calibration = "Rabi"
        self._user_variables_checks(calibration, scan_variables, iterations)
        self._user_variables_init(calibration, scan_variables, iterations, cooldown_time)

    def set_T1(self, scan_variables, iterations=1, cooldown_time=None):
        """Performs a T1 measurement by sending a pi-pulse to the qubit and measuring the resonator response after a varying time.
        The available scanning parameters are:
        'duration' (wait time in clock cycles (4ns), must be an integer).

        :param scan_variables: variables to be scanned. Must be an array of tuples [('variable', [values])].
        :param iterations: number of averaging iterations. Must be a positive Python integer.
        :param cooldown_time: resonator or qubit cooldown time in clock cycles (4ns). Must be an integer.
        :return: None
        """
        calibration = "T1"
        self._user_variables_checks(calibration, scan_variables, iterations)
        self._user_variables_init(calibration, scan_variables, iterations, cooldown_time)
        if (len(scan_variables) > 1) or (scan_variables[0][0] != "duration"):
            raise Exception("Only 'duration' scans are available for T1 calibration.")

    def set_ramsey(self, scan_variables, iterations=1, cooldown_time=None, idle_time=None):
        """Performs a Ramsey-like measurement (pi/2 - idle time - pi/2).
        The available scanning parameters are:
        'duration' (idle time in clock cycles unit (4 ns)),
        'frequency' (pulse frequency in Hz, must be an integer).

        :param scan_variables: variables to be scanned. Must be an array of tuples [('variable', [values])]. 1D and 2D scans are available.
        :param iterations: number of averaging iterations. Must be a positive Python integer.
        :param cooldown_time: resonator or qubit cooldown time in clock cycles (4ns). Must be an integer.
        :param idle_time: Ramsey sequence idle time in clock cycles unit (4 ns). Must be an integer larger than 4.
        :return: None
        """
        calibration = "Ramsey"
        self._user_variables_checks(calibration, scan_variables, iterations)
        self._user_variables_init(calibration, scan_variables, iterations, cooldown_time)
        # Check idle time
        if idle_time is None or (idle_time >= 4 and type(idle_time) is int):
            self.user_specifics[calibration]["idle_time"] = idle_time
        elif cooldown_time == 0:
            self.user_specifics[calibration]["cooldown_time"] = None
        else:
            raise Exception("idle_time must be 0 or a python positive integer larger than 4.")

    def set_raw_traces(self, scan_variables=(), iterations=1, cooldown_time=None):
        """Acquires the raw and averaged traces from the two inputs of the OPX.

        :param scan_variables: None
        :param iterations: number of averaging iterations. Must be a positive Python integer.
        :param cooldown_time: resonator or qubit cooldown time in clock cycles (4ns). Must be an integer.
        :return: None
        """
        calibration = "raw_traces"
        self._user_variables_checks(calibration, scan_variables, iterations)
        self._user_variables_init(calibration, scan_variables, iterations, cooldown_time)

    def set_time_of_flight(self, scan_variables=(), iterations=1, cooldown_time=None, threshold=10 / 4096):
        """Measures the time of flight and offsets of the two inputs of the OPX.

        :param scan_variables: None
        :param iterations: number of averaging iterations. Must be a positive Python integer.
        :param cooldown_time: resonator or qubit cooldown time in clock cycles (4ns). Must be an integer.
        :param threshold: detection threshold for time of flight estimation in V/ns.
        :return: None
        """
        calibration = "time_of_flight"
        self._user_variables_checks(calibration, scan_variables, iterations)
        self._user_variables_init(calibration, scan_variables, iterations, cooldown_time)
        self.user_specifics[calibration]["threshold"] = threshold

    def run_calibrations(self, quantum_machine, plot=None, fits=False, plot_options=None):
        """Execute the implemented calibrations.

        :param quantum_machine:
        :param plot: Can be either 'full' for plotting the data after the execution or 'live' for live plotting.
        :param fits: To be implemented
        :param plot_options: Python dictionary containing options to customize the plot (ex: options = {"fontsize": 14, "color": "b", "marker":'o', "linewidth":0, "figsize": (12, 15)})
        :return: None
        """
        self.qm = quantum_machine
        self._check_plot_options(plot_options)

        for calib in self.calibration_list:
            print(f"The calibration: {calib} is in progress...")
            if calib == "Rabi":
                self._rabi(
                    scan_dimension=len(self.scan_var[calib]) - 1,
                    plot=plot,
                    fits=fits,
                )
            elif calib == "T1":
                self._T1(
                    scan_dimension=len(self.scan_var[calib]) - 1,
                    plot=plot,
                    fits=fits,
                )
            elif calib == "Ramsey":
                self._ramsey(
                    scan_dimension=len(self.scan_var[calib]) - 1,
                    plot=plot,
                    fits=fits,
                )
            elif calib == "Resonator_spec":
                self._resonator_spec(
                    scan_dimension=len(self.scan_var[calib]) - 1,
                    plot=plot,
                    fits=fits,
                )
            elif calib == "raw_traces":
                self._raw_traces(plot=plot)
            elif calib == "time_of_flight":
                self._time_of_flight(plot=plot)

    def simulate_calibrations(self, machine, simulation_duration):
        """Simulate the implemented calibrations.

        :param machine: either a QuantumMachine or a QuantumMachinesManager instance that will be used to run the simulation.
        :param simulation_duration: duration of the simulation in clock cycle unit (4ns)
        :return: None
        """
        self.simulation = True
        for calib in self.calibration_list:
            prog = None
            if calib == "Rabi":
                prog = self._rabi(scan_dimension=len(self.scan_var[calib]) - 1)
            if calib == "T1":
                prog = self._T1(scan_dimension=len(self.scan_var[calib]) - 1)
            elif calib == "Ramsey":
                prog = self._ramsey(scan_dimension=len(self.scan_var[calib]) - 1)
            elif calib == "Resonator_spec":
                prog = self._resonator_spec(scan_dimension=len(self.scan_var[calib]) - 1)
            elif calib == "raw_traces":
                prog = self._raw_traces()
            elif calib == "time_of_flight":
                prog = self._time_of_flight()

            simulation_config = SimulationConfig(duration=int(simulation_duration))
            if isinstance(machine, QuantumMachinesManager):
                job = machine.simulate(self.config, prog, simulation_config)
            elif isinstance(machine, QuantumMachine):
                job = machine.simulate(prog, simulation_config)
            else:
                raise Exception("'machine' must be either a QuantumMachine or a QuantumMachinesManager instance.")
            plt.figure()
            job.get_simulated_samples().con1.plot()
        self.simulation = False

    def _resonator_spec(self, scan_dimension, plot=False, fits=False):
        calib = "Resonator_spec"
        if scan_dimension == 1:
            with program() as resonator_spec_1D:
                # Get the QUA variables for the 2D scan + averaging
                n, var1, var1_scan, var1_name = self._get_qua_variables(calib)
                if var1_name == "amplitude" and self.flux_line_elmt is None:
                    raise Exception(
                        "Amplitude scan of the resonator spectroscopy is only implemented for flux tunable transmons so please provide the flux line element."
                    )
                elif var1_name != "frequency" or (var1_name == "amplitude" and self.flux_line_elmt is None):
                    raise Exception("Resonator spectroscopy is only implemented for frequency scan.")

                I = declare(fixed)
                Q = declare(fixed)
                I_st = declare_stream()
                Q_st = declare_stream()
                n_st = declare_stream()

                with for_(n, 0, n < self.scan_var[calib]["averaging"], n + 1):
                    with for_(*from_array(var1, var1_scan)):
                        if self.flux_line_elmt is not None:
                            if var1_name == "amplitude":
                                play(self.flux_line_op * amp(var1), self.flux_line_elmt)
                            else:
                                play(self.flux_line_op, self.flux_line_elmt)
                        # Update the resonator frequency
                        if var1_name == "frequency":
                            update_frequency(self.readout_elmt, var1)
                        # Align, measure and save data
                        self._align_measure_save(I, Q, I_st, Q_st, calib, align_flag=False)
                    save(n, n_st)

                with stream_processing():
                    I_st.buffer(len(var1_scan)).average().save("I")
                    Q_st.buffer(len(var1_scan)).average().save("Q")
                    n_st.save("iteration")

            if self.simulation:
                return resonator_spec_1D
            job = self.qm.execute(resonator_spec_1D)

            self._get_IQ_data_and_plot(job, calib, plot, scan_dimension)

        elif scan_dimension == 2:
            if self.flux_line_elmt is None:
                raise Exception(
                    "2D scan with 'amplitude' or 'duration' for the resonator spectroscopy is only implemented for flux tunable transmons so please provide the flux line element."
                )
            with program() as resonator_spec_2D:
                # Get the QUA variables for the 2D scan + averaging
                (
                    n,
                    var1,
                    var1_scan,
                    var1_name,
                    var2,
                    var2_scan,
                    var2_name,
                ) = self._get_qua_variables(calib)

                I = declare(fixed)
                Q = declare(fixed)
                I_st = declare_stream()
                Q_st = declare_stream()
                n_st = declare_stream()
                with for_(n, 0, n < self.scan_var[calib]["averaging"], n + 1):
                    with for_(*from_array(var2, var2_scan)):
                        with for_(*from_array(var1, var1_scan)):
                            if var1_name == "amplitude":
                                play(self.flux_line_op * amp(var1), self.flux_line_elmt)
                            elif var2_name == "amplitude":
                                play(self.flux_line_op * amp(var2), self.flux_line_elmt)
                            else:
                                play(self.flux_line_op, self.flux_line_elmt)

                            # Update the resonator frequency
                            if var1_name == "frequency":
                                update_frequency(self.readout_elmt, var1)
                            elif var2_name == "frequency":
                                update_frequency(self.readout_elmt, var2)
                            # Align, measure and save data
                            self._align_measure_save(I, Q, I_st, Q_st, calib, align_flag=False)
                    save(n, n_st)

                with stream_processing():
                    I_st.buffer(len(var1_scan)).buffer(len(var2_scan)).average().save("I")
                    Q_st.buffer(len(var1_scan)).buffer(len(var2_scan)).average().save("Q")
                    n_st.save("iteration")

            if self.simulation:
                return resonator_spec_2D
            job = self.qm.execute(resonator_spec_2D)
            self._get_IQ_data_and_plot(job, calib, plot, scan_dimension)

    def _rabi(self, scan_dimension, plot=False, fits=False):
        calib = "Rabi"
        if scan_dimension == 1:
            with program() as prog_rabi_1D:
                # Get the QUA variables for the 2D scan + averaging
                n, var1, var1_scan, var1_name = self._get_qua_variables(calib)
                I = declare(fixed)
                Q = declare(fixed)
                I_st = declare_stream()
                Q_st = declare_stream()
                n_st = declare_stream()

                with for_(n, 0, n < self.scan_var[calib]["averaging"], n + 1):
                    with for_(*from_array(var1, var1_scan)):

                        # Update the resonator frequency
                        if var1_name == "frequency":
                            update_frequency(self.qubit_elmt, var1)
                        # Adjust the pulse amplitude and/or duration
                        if var1_name == "amplitude":
                            play(self.qubit_op * amp(var1), self.qubit_elmt)
                        elif var1_name == "duration":
                            play(self.qubit_op, self.qubit_elmt, duration=var1)
                        # Align, measure and save data
                        self._align_measure_save(I, Q, I_st, Q_st, calib)
                    save(n, n_st)

                with stream_processing():
                    I_st.buffer(len(var1_scan)).average().save("I")
                    Q_st.buffer(len(var1_scan)).average().save("Q")
                    n_st.save("iteration")

            if self.simulation:
                return prog_rabi_1D
            job = self.qm.execute(prog_rabi_1D)

            self._get_IQ_data_and_plot(job, calib, plot, scan_dimension)

        elif scan_dimension == 2:
            with program() as prog_rabi_2D:
                # Get the QUA variables for the 2D scan + averaging
                (
                    n,
                    var1,
                    var1_scan,
                    var1_name,
                    var2,
                    var2_scan,
                    var2_name,
                ) = self._get_qua_variables(calib)

                I = declare(fixed)
                Q = declare(fixed)
                I_st = declare_stream()
                Q_st = declare_stream()
                n_st = declare_stream()
                with for_(n, 0, n < self.scan_var[calib]["averaging"], n + 1):
                    with for_(*from_array(var2, var2_scan)):
                        with for_(*from_array(var1, var1_scan)):
                            # Update the resonator frequency
                            if var1_name == "frequency":
                                update_frequency(self.qubit_elmt, var1)
                            elif var2_name == "frequency":
                                update_frequency(self.qubit_elmt, var2)

                            # Adjust the pulse amplitude and/or duration
                            if var1_name == "amplitude":
                                if var2_name == "duration":
                                    play(
                                        self.qubit_op * amp(var1),
                                        self.qubit_elmt,
                                        duration=var2,
                                    )
                                else:
                                    play(self.qubit_op * amp(var1), self.qubit_elmt)
                            elif var2_name == "amplitude":
                                if var1_name == "duration":
                                    play(
                                        self.qubit_op * amp(var2),
                                        self.qubit_elmt,
                                        duration=var1,
                                    )
                                else:
                                    play(self.qubit_op * amp(var2), self.qubit_elmt)
                            else:
                                if var1_name == "duration":
                                    play(self.qubit_op, self.qubit_elmt, duration=var1)
                                else:
                                    play(self.qubit_op, self.qubit_elmt, duration=var2)

                            # Align, measure and save data
                            self._align_measure_save(I, Q, I_st, Q_st, calib)
                    save(n, n_st)

                with stream_processing():
                    I_st.buffer(len(var1_scan)).buffer(len(var2_scan)).average().save("I")
                    Q_st.buffer(len(var1_scan)).buffer(len(var2_scan)).average().save("Q")
                    n_st.save("iteration")

            if self.simulation:
                return prog_rabi_2D
            job = self.qm.execute(prog_rabi_2D)

            self._get_IQ_data_and_plot(job, calib, plot, scan_dimension)

    def _T1(self, scan_dimension, plot=False, fits=False):
        calib = "T1"
        with program() as prog_T1:
            # Get the QUA variables for the 2D scan + averaging
            n, var1, var1_scan, var1_name = self._get_qua_variables(calib)
            I = declare(fixed)
            Q = declare(fixed)
            I_st = declare_stream()
            Q_st = declare_stream()
            n_st = declare_stream()

            with for_(n, 0, n < self.scan_var[calib]["averaging"], n + 1):
                with for_(*from_array(var1, var1_scan)):
                    # Play pi pulse
                    play(self.qubit_op, self.qubit_elmt)
                    align(self.qubit_elmt, self.readout_elmt)
                    # scan wait duration
                    wait(var1, self.readout_elmt)
                    # Align, measure and save data
                    self._align_measure_save(I, Q, I_st, Q_st, calib)
                save(n, n_st)

            with stream_processing():
                I_st.buffer(len(var1_scan)).average().save("I")
                Q_st.buffer(len(var1_scan)).average().save("Q")
                n_st.save("iteration")

        if self.simulation:
            return prog_T1
        job = self.qm.execute(prog_T1)

        self._get_IQ_data_and_plot(job, calib, plot, scan_dimension)

    def _ramsey(self, scan_dimension, plot=False, fits=False):
        calib = "Ramsey"
        if scan_dimension == 1:
            with program() as prog_ramsey_1D:
                # Get the QUA variables for the 2D scan + averaging
                n, var1, var1_scan, var1_name = self._get_qua_variables(calib)
                I = declare(fixed)
                Q = declare(fixed)
                I_st = declare_stream()
                Q_st = declare_stream()
                n_st = declare_stream()

                with for_(n, 0, n < self.scan_var[calib]["averaging"], n + 1):
                    with for_(*from_array(var1, var1_scan)):
                        # Update the resonator frequency
                        if var1_name == "frequency":
                            update_frequency(self.qubit_elmt, var1)
                        # First pi/2 pulse
                        play(self.qubit_op, self.qubit_elmt)
                        # Adjust the idle time
                        if var1_name == "duration":
                            wait(var1, self.qubit_elmt)
                        elif self.user_specifics[calib]["idle_time"] is not None:
                            wait(self.user_specifics[calib]["idle_time"], self.qubit_elmt)
                        # Second pi/2 pulse
                        play(self.qubit_op, self.qubit_elmt)
                        # Align, measure and save data
                        self._align_measure_save(I, Q, I_st, Q_st, calib)
                    save(n, n_st)

                with stream_processing():
                    I_st.buffer(len(var1_scan)).average().save("I")
                    Q_st.buffer(len(var1_scan)).average().save("Q")
                    n_st.save("iteration")

            if self.simulation:
                return prog_ramsey_1D
            job = self.qm.execute(prog_ramsey_1D)

            self._get_IQ_data_and_plot(job, calib, plot, scan_dimension)

        elif scan_dimension == 2:
            with program() as prog_ramsey_2D:

                # Get the QUA variables for the 2D scan + averaging
                (
                    n,
                    var1,
                    var1_scan,
                    var1_name,
                    var2,
                    var2_scan,
                    var2_name,
                ) = self._get_qua_variables(calib)

                I = declare(fixed)
                Q = declare(fixed)
                I_st = declare_stream()
                Q_st = declare_stream()
                n_st = declare_stream()

                with for_(n, 0, n < self.scan_var[calib]["averaging"], n + 1):
                    with for_(*from_array(var2, var2_scan)):
                        with for_(*from_array(var1, var1_scan)):
                            # Update the resonator frequency
                            if var1_name == "frequency":
                                update_frequency(self.qubit_elmt, var1)
                            elif var2_name == "frequency":
                                update_frequency(self.qubit_elmt, var2)
                            # First pi/2 pulse
                            play(self.qubit_op, self.qubit_elmt)
                            # Adjust the idle time
                            if var1_name == "duration":
                                wait(var1, self.qubit_elmt)
                            elif var2_name == "duration":
                                wait(var2, self.qubit_elmt)
                            elif self.user_specifics[calib]["idle_time"] is not None:
                                wait(
                                    self.user_specifics[calib]["idle_time"],
                                    self.qubit_elmt,
                                )
                            # Second pi/2 pulse
                            play(self.qubit_op, self.qubit_elmt)

                            # Align, measure and save data
                            self._align_measure_save(I, Q, I_st, Q_st, calib)
                    save(n, n_st)

                with stream_processing():
                    I_st.buffer(len(var1_scan)).buffer(len(var2_scan)).average().save("I")
                    Q_st.buffer(len(var1_scan)).buffer(len(var2_scan)).average().save("Q")
                    n_st.save("iteration")

            if self.simulation:
                return prog_ramsey_2D
            job = self.qm.execute(prog_ramsey_2D)

            self._get_IQ_data_and_plot(job, calib, plot, scan_dimension)

    def _raw_traces(self, plot=False):
        calib = "raw_traces"
        with program() as raw_trace_prog:
            n = declare(int)
            adc_st = declare_stream(adc_trace=True)

            with for_(n, 0, n < self.scan_var[calib]["averaging"], n + 1):
                reset_phase(self.readout_elmt)
                measure(self.readout_op, self.readout_elmt, adc_st)
                if self.user_specifics[calib]["cooldown_time"] is not None:
                    wait(
                        int(self.user_specifics[calib]["cooldown_time"]),
                        self.readout_elmt,
                    )

            with stream_processing():
                # Will save average:
                adc_st.input1().average().save("adc1")
                adc_st.input2().average().save("adc2")
                # Will save only last run:
                adc_st.input1().save("adc1_single_run")
                adc_st.input2().save("adc2_single_run")

        if self.simulation:
            return raw_trace_prog
        job = self.qm.execute(raw_trace_prog)

        res_handles = job.result_handles

        if plot == "live":
            adc1_handles = res_handles.get("adc1_single_run")
            adc2_handles = res_handles.get("adc2_single_run")
            adc1_av_handles = res_handles.get("adc1")
            adc2_av_handles = res_handles.get("adc2")
            adc1_handles.wait_for_values(1)
            adc2_handles.wait_for_values(1)
            adc1_av_handles.wait_for_values(1)
            adc2_av_handles.wait_for_values(1)

            # Live plotting
            fig = plt.figure(figsize=self.plot_options["figsize"])
            interrupt_on_close(fig, job)  # Interrupts the job when closing the figure
            while res_handles.is_processing():
                adc1 = u.raw2volts(res_handles.get("adc1").fetch_all())
                adc2 = u.raw2volts(res_handles.get("adc2").fetch_all())
                adc1_single_run = u.raw2volts(res_handles.get("adc1_single_run").fetch_all())
                adc2_single_run = u.raw2volts(res_handles.get("adc2_single_run").fetch_all())

                self.results[calib]["adc1"] = {"raw": adc1_single_run, "averaged": adc1}
                self.results[calib]["adc2"] = {"raw": adc2_single_run, "averaged": adc2}

                plt.subplot(121)
                plt.cla()
                plt.title(
                    "Single run (Check ADCs saturation)",
                    fontsize=self.plot_options["fontsize"] + 2,
                )
                plt.plot(adc1_single_run, "b")
                plt.plot(adc2_single_run, "r")
                plt.xlabel("Time [ns]", fontsize=self.plot_options["fontsize"])
                plt.ylabel("Signal amplitude [V]", fontsize=self.plot_options["fontsize"])
                plt.subplot(122)
                plt.cla()
                plt.title("Averaged run", fontsize=self.plot_options["fontsize"] + 2)
                plt.plot(adc1, "b")
                plt.plot(adc2, "r")
                plt.xlabel("Time [ns]", fontsize=self.plot_options["fontsize"])
                plt.ylabel("Signal amplitude [V]", fontsize=self.plot_options["fontsize"])
                plt.pause(0.01)
        else:
            res_handles.wait_for_all_values()
            adc1 = u.raw2volts(res_handles.get("adc1").fetch_all())
            adc2 = u.raw2volts(res_handles.get("adc2").fetch_all())
            adc1_single_run = u.raw2volts(res_handles.get("adc1_single_run").fetch_all())
            adc2_single_run = u.raw2volts(res_handles.get("adc2_single_run").fetch_all())

            self.results[calib]["adc1"] = {"raw": adc1_single_run, "averaged": adc1}
            self.results[calib]["adc2"] = {"raw": adc2_single_run, "averaged": adc2}

            if plot == "full":
                plt.figure(figsize=self.plot_options["figsize"])
                plt.subplot(121)
                plt.title(
                    "Single run (Check ADCs saturation)",
                    fontsize=self.plot_options["fontsize"] + 2,
                )
                plt.plot(adc1_single_run, "b")
                plt.plot(adc2_single_run, "r")
                plt.xlabel("Time [ns]", fontsize=self.plot_options["fontsize"])
                plt.ylabel("Signal amplitude [V]", fontsize=self.plot_options["fontsize"])
                plt.subplot(122)
                plt.title("Averaged run", fontsize=self.plot_options["fontsize"] + 2)
                plt.plot(adc1, "b")
                plt.plot(adc2, "r")
                plt.xlabel("Time [ns]", fontsize=self.plot_options["fontsize"])
                plt.ylabel("Signal amplitude [V]", fontsize=self.plot_options["fontsize"])

    def _time_of_flight(self, plot=False):
        calib = "time_of_flight"
        with program() as tof_prog:
            n = declare(int)
            adc_st = declare_stream(adc_trace=True)
            with for_(n, 0, n < self.scan_var[calib]["averaging"], n + 1):
                reset_phase(self.readout_elmt)
                measure(self.readout_op, self.readout_elmt, adc_st)
                if self.user_specifics[calib]["cooldown_time"] is not None:
                    wait(
                        int(self.user_specifics[calib]["cooldown_time"]),
                        self.readout_elmt,
                    )

            with stream_processing():
                # Will save average:
                if "out1" in self.config["elements"][self.readout_elmt]["outputs"]:
                    adc_st.input1().average().save("adc1")
                if "out2" in self.config["elements"][self.readout_elmt]["outputs"]:
                    adc_st.input2().average().save("adc2")
        if self.simulation:
            return tof_prog
        job = self.qm.execute(tof_prog)

        res_handles = job.result_handles
        if plot == "live":
            raise Exception("Live plotting is not available for time of flight calibration")
        else:
            adc1 = None
            adc2 = None
            tof1 = None
            tof2 = None
            res_handles.wait_for_all_values()
            if "out1" in self.config["elements"][self.readout_elmt]["outputs"]:
                adc1 = u.raw2volts(res_handles.get("adc1").fetch_all())
                self.results[calib]["adc1"] = adc1
                # Find the pulse edge and derive tof
                if len(np.where(np.abs(np.diff(adc1)) > self.user_specifics[calib]["threshold"])[0]) > 0:
                    tof1 = np.min(np.where(np.abs(np.diff(adc1)) > self.user_specifics[calib]["threshold"]))
                    print(f"TOF to add = {tof1} ns for out1")
                else:
                    tof1 = None
                    print("No pulse detected")
            if "out2" in self.config["elements"][self.readout_elmt]["outputs"]:
                adc2 = u.raw2volts(res_handles.get("adc2").fetch_all())
                self.results[calib]["adc2"] = adc2
                # Find the pulse edge and derive tof
                if len(np.where(np.abs(np.diff(adc2)) > self.user_specifics[calib]["threshold"])[0]) > 0:
                    tof2 = np.min(np.where(np.abs(np.diff(adc2)) > self.user_specifics[calib]["threshold"]))
                    print(f"TOF to add = {tof2} ns for out2")
                else:
                    tof2 = None
                    print("No pulse detected")

            if plot == "full":
                plt.figure(figsize=self.plot_options["figsize"])
                plt.title("Time of flight", fontsize=self.plot_options["fontsize"] + 2)
                if "out1" in self.config["elements"][self.readout_elmt]["outputs"]:
                    plt.plot(
                        adc1,
                        "b",
                        label=f"TOF to add: {tof1} ns and offset to remove: {np.mean(adc1[:tof1]) * 1000:.4f} mV",
                    )
                    if tof1 is not None:
                        plt.axvline(x=tof1, color="b", linestyle=":")
                    plt.axhline(y=np.mean(adc1[:tof1]), color="b", linestyle="--")
                if "out2" in self.config["elements"][self.readout_elmt]["outputs"]:
                    plt.plot(
                        adc2,
                        "r",
                        label=f"TOF to add: {tof2} ns and offset to remove: {np.mean(adc2[:tof2]) * 1000:.4f} mV",
                    )
                    if tof2 is not None:
                        plt.axvline(x=tof2, color="r", linestyle=":")
                    plt.axhline(y=np.mean(adc2[:tof2]), color="r", linestyle="--")
                plt.xlabel("Time [ns]", fontsize=self.plot_options["fontsize"])
                plt.ylabel("Signal amplitude [V]", fontsize=self.plot_options["fontsize"])
                plt.legend()



class MultiObjectTempAttrUpdater:
    def __init__(self, objects, **temp_attrs):
        """
        Initialize the MultiObjectTempAttrUpdater context manager.

        Parameters:
        - objects: A list of objects whose attributes will be temporarily updated.
        - temp_attrs: The temporary attributes and their values to be set.
        """
        self.objects = objects
        self.temp_attrs = temp_attrs
        self.original_attrs = []

    def __enter__(self):
        """
        Enter the runtime context related to this object.
        Temporarily set the objects' attributes to the new values.
        """
        # Store original attributes
        for obj in self.objects:
            original_attrs = {}
            for attr, value in self.temp_attrs.items():
                if hasattr(obj, attr):
                    original_attrs[attr] = getattr(obj, attr)
                else:
                    original_attrs[attr] = None

                # Set new temporary attribute value
                setattr(obj, attr, value)

            self.original_attrs.append((obj, original_attrs))

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        Exit the runtime context.
        Restore the objects' original attributes.
        """
        for obj, original_attrs in self.original_attrs:
            for attr, value in original_attrs.items():
                setattr(obj, attr, value)



import warnings
from typing import Tuple, List
import numpy as np
import scipy.signal as sig
from warnings import warn
from enum import Enum


class QOPVersion(Enum):
    """A dictionary with the filter limitations for the QOP versions"""

    NONE = {
        "feedforward_max": np.inf,
        "feedback_max": np.inf,
        "feedforward_length": lambda feedback_len: np.inf,
    }
    QOP222 = {
        "feedforward_max": 2 - 2**-16,
        "feedback_max": 1 - 2**-20,
        "feedforward_length": lambda feedback_len: 44 - 7 * feedback_len,
    }
    QOP220 = {
        "feedforward_max": 2 - 2**-16,
        "feedback_max": 1 - 2**-20,
        "feedforward_length": lambda feedback_len: 44 - 7 * feedback_len,
    }

    @classmethod
    def get_latest(cls):
        """Return the latest QOP version."""
        return cls.QOP222

    @classmethod
    def get_options(cls):
        """Return the list of implemented QOP versions"""
        return [x.name for x in cls]


def calc_filter_taps(
    fir: List[float] = None,
    exponential: List[Tuple[float, float]] = None,
    highpass: List[float] = None,
    bounce: List[Tuple[float, float]] = None,
    delay: float = None,
    Ts: float = 1,
    qop_version: QOPVersion = QOPVersion.get_latest(),
) -> Tuple[List[float], List[float]]:
    """
    Calculate the best FIR and IIR filter taps for a system with any combination of FIR corrections, exponential
    corrections (undershoot or overshoot), high pass compensation, reflections (bounce corrections) and a needed delay on the line.
    The OPX has hardware constraints that may limit the filter implementation and this is why the running QOP version can be specified as an enum of the class QOPVersion.
    The possible options are returned by the `QOPVersion.get_options()` method and the default value is given by `QOPVersion.get_latest()`.

    Args:
        fir: A list of the needed FIR taps. These would be convoluted with whatever other FIR taps
            are required by the other filters.
        exponential: A list of tuples (A, tau), each tuple represents an exponential decay of the shape
            `1 + A * exp(-t/tau)`. `tau` is in ns. Each exponential correction requires 1 IIR tap and 2 FIR taps.
        highpass: A list of taus, each tau represents a highpass decay of the shape `exp(-t/tau)`.
            `tau` is in ns. Each highpass correction requires 1 IIR tap and 2 FIR taps.
        bounce: A list of tuples (a, tau), each tuple represents a reflection of amplitude `a` happening at time `tau`.
            `tau` is in ns. Note, if `tau` is not a multiple of the sampling rate, multiple FIR taps will be created.
            If `tau` is smaller than 5 taps, accuracy might be lost.
        delay: A global delay to apply using the FIR filters.
            `delay` is in ns. Note, if `delay` is not a multiple of the sampling rate, multiple FIR taps will be
            created. If `delay` is smaller than 5 taps, accuracy might be lost.
        Ts: The sampling rate (in ns) of the system and filter taps.
        qop_version: running QOP version used to format the taps according to the corresponding hardware limitations (ex: QOPVersion.QOP222).
    Returns:
        A tuple of two lists.
        The first is a list of FIR (feedforward) taps starting at 0 and spaced `Ts` apart.
        The second is a list of IIR (feedback) taps.
    """
    feedforward_taps = np.array([1.0])
    feedback_taps = np.array([])

    if highpass is not None:
        feedforward_taps, feedback_taps = _iir_correction(highpass, "highpass", feedforward_taps, feedback_taps, Ts)

    if exponential is not None:
        feedforward_taps, feedback_taps = _iir_correction(
            exponential, "exponential", feedforward_taps, feedback_taps, Ts
        )

    if fir is not None:
        feedforward_taps = np.convolve(feedforward_taps, fir)

    if bounce is not None or delay is not None:
        feedforward_taps = bounce_and_delay_correction(bounce, delay, feedforward_taps, Ts, qop_version=QOPVersion.NONE)

    return _check_hardware_limitation(qop_version, feedforward_taps, list(feedback_taps))


def multi_exponential_decay(x, amplitudes, decay_times, s=1):
    """
    Function representing the multi-exponential decay defined as
    s * (1 + a1 * np.exp(-x / t1) + a2 * np.exp(-x / t2) + ... + an * np.exp(-x / tn)).

    :param x: numpy array for the time vector in ns
    :param amplitudes: list of floats for the exponential amplitudes [a1, a2, ..., an]
    :param decay_times: list of floats for the exponential decay times in ns [t1, t2, ..., tn]
    :param s: float for the offset of the exponential decay (default: 1)
    :return: numpy array for the multi-exponential decay

    examples:

    # Single exponential decay
    multi_exponential_decay(x, [a], [t], s)

    # Dual exponential decay
    multi_exponential_decay(x, [a1, a2], [t1, t2], s)

    # Triple exponential decay
    multi_exponential_decay(x, [a1, a2, a3], [t1, t2, t3], s)
    """
    decay = s * (1 + sum(a * np.exp(-x / t) for a, t in zip(amplitudes, decay_times)))
    return decay


def exponential_decay(x, a, t):
    """
    Function representing the exponential decay defined as 1 + a * np.exp(-x / t).

    :param x: numpy array for the time vector in ns
    :param a: float for the exponential amplitude
    :param t: float for the exponential decay time in ns
    :return: numpy array for the exponential decay
    """
    return multi_exponential_decay(x, [a], [t])


def high_pass_exponential(x, t):
    """Function representing the exponential decay defined as np.exp(-x / t).

    :param x: numpy array for the time vector in ns
    :param t: float for the exponential decay time in ns
    :return: numpy array for the exponential decay
    """
    return np.exp(-x / t)


def single_exponential_correction(
    A: float, tau: float, Ts: float = 1, qop_version: QOPVersion = QOPVersion.get_latest()
):
    """
    Calculate the best FIR and IIR filter taps to correct for an exponential decay (undershoot or overshoot) of the shape
    `1 + A * exp(-t/tau)`.
    The OPX has hardware constraints that may limit the filter implementation and this is why the running QOP version can be specified as an enum of the class QOPVersion.
    The possible options are returned by the `QOPVersion.get_options()` method and the default value is given by `QOPVersion.get_latest()`.

    Args:
        A: The exponential decay pre-factor.
        tau: The time constant for the exponential decay, given in ns.
        Ts: The sampling rate (in ns) of the system and filter taps.
        qop_version: running QOP version used to format the taps according to the corresponding hardware limitations (ex: QOPVersion.QOP222).
    Returns:
        A tuple of two items.
        The first is a list of 2 FIR (feedforward) taps starting at 0 and spaced `Ts` apart.
        The second is a single IIR (feedback) tap.
    """
    tau *= 1e-9
    Ts *= 1e-9
    k1 = Ts + 2 * tau * (A + 1)
    k2 = Ts - 2 * tau * (A + 1)
    c1 = Ts + 2 * tau
    c2 = Ts - 2 * tau
    feedback_tap = [-k2 / k1]
    feedforward_taps = list(np.array([c1, c2]) / k1)
    return _check_hardware_limitation(qop_version, feedforward_taps, feedback_tap)


def highpass_correction(tau: float, Ts: float = 1, qop_version: QOPVersion = QOPVersion.get_latest()):
    """
    Calculate the best FIR and IIR filter taps to correct for a highpass decay (HPF) of the shape `exp(-t/tau)`.
    The OPX has hardware constraints that may limit the filter implementation and this is why the running QOP version can be specified as an enum of the class QOPVersion.
    The possible options are returned by the `QOPVersion.get_options()` method and the default value is given by `QOPVersion.get_latest()`.

    Args:
        tau: The time constant for the exponential decay, given in ns.
        Ts: The sampling rate (in ns) of the system and filter taps.
        qop_version: running QOP version used to format the taps according to the corresponding hardware limitations (ex: QOPVersion.QOP222).
    Returns:
        A tuple of two items.
        The first is a list of 2 FIR (feedforward) taps starting at 0 and spaced `Ts` apart.
        The second is a single IIR (feedback) tap.
    """
    Ts *= 1e-9
    flt = sig.butter(1, np.array([1 / tau / Ts]), btype="highpass", analog=True)
    ahp2, bhp2 = sig.bilinear(flt[1], flt[0], 1e9)
    feedforward_taps = list(np.array([ahp2[0], ahp2[1]]))
    feedback_tap = [bhp2[0]]
    return _check_hardware_limitation(qop_version, feedforward_taps, feedback_tap)


def bounce_and_delay_correction(
    bounce_values: list = (),
    delay: float = 0,
    feedforward_taps: list = (1.0,),
    Ts: float = 1,
    qop_version: QOPVersion = QOPVersion.get_latest(),
):
    """
    Calculate the FIR filter taps to correct for reflections (bounce corrections) and to add a delay.
    The OPX has hardware constraints that may limit the filter implementation and this is why the running QOP version can be specified as an enum of the class QOPVersion.
    The possible options are returned by the `QOPVersion.get_options()` method and the default value is given by `QOPVersion.get_latest()`.

    Args:
        bounce_values: A list of tuples (a, tau), each tuple represents a reflection of amplitude `a` happening at time
            `tau`. `tau` is in ns. Note, if `tau` is not a multiple of the sampling rate, multiple FIR taps will be
            created. If `tau` is smaller than 5 taps, accuracy might be lost.
        delay: A global delay to apply using the FIR filters.
            `delay` is in ns. Note, if `delay` is not a multiple of the sampling rate, multiple FIR taps will be
            created. If `delay` is smaller than 5 taps, accuracy might be lost.
        feedforward_taps: Existing FIR (feedforward) taps to be convoluted with the resulting taps.
        Ts: The sampling rate (in ns) of the system and filter taps.
        qop_version: running QOP version used to format the taps according to the corresponding hardware limitations (ex: QOPVersion.QOP222).
    Returns:
        A list of FIR (feedforward) taps starting at 0 and spaced `Ts` apart.
    """
    if bounce_values is None:
        bounce_values = []
    if delay is None:
        delay = 0

    if feedforward_taps is None:
        feedforward_taps = [1.0]
    n_extra = 10
    n_taps = 101
    long_taps_x = np.linspace((0 - n_extra) * Ts, (n_taps + n_extra) * Ts, n_taps + 1 + 2 * n_extra)[0:-1]
    feedforward_taps_x = np.linspace(0, (len(feedforward_taps) - 1) * Ts, len(feedforward_taps))

    delay_taps = _get_coefficients_for_delay(delay, long_taps_x, Ts)

    feedforward_taps = np.convolve(feedforward_taps, delay_taps)
    feedforward_taps_x = np.linspace(
        min(feedforward_taps_x) + min(long_taps_x),
        max(feedforward_taps_x) + max(long_taps_x),
        len(feedforward_taps),
    )
    for i, (a, tau) in enumerate(bounce_values):
        bounce_taps = -a * _get_coefficients_for_delay(tau, long_taps_x, Ts)
        bounce_taps[n_extra] += 1
        feedforward_taps = np.convolve(feedforward_taps, bounce_taps)
        feedforward_taps_x = np.linspace(
            min(feedforward_taps_x) + min(long_taps_x),
            max(feedforward_taps_x) + max(long_taps_x),
            len(feedforward_taps),
        )

    feedforward_taps = _round_taps_close_to_zero(feedforward_taps)
    index_start = np.nonzero(feedforward_taps_x == 0)[0][0]
    index_end = np.nonzero(feedforward_taps)[0][-1] + 1
    final_taps = feedforward_taps[index_start:index_end]

    return _check_hardware_limitation(qop_version, final_taps, [])[0]


def _iir_correction(values, filter_type, feedforward_taps, feedback_taps, Ts=1.0):
    b = np.zeros((2, len(values)))
    feedback_taps = np.append(np.zeros(len(values)), feedback_taps)

    if filter_type == "highpass":
        for i, tau in enumerate(values):
            b[:, i], [feedback_taps[i]] = highpass_correction(tau, Ts, qop_version=QOPVersion.NONE)
    elif filter_type == "exponential":
        for i, (A, tau) in enumerate(values):
            b[:, i], [feedback_taps[i]] = single_exponential_correction(A, tau, Ts, qop_version=QOPVersion.NONE)
    else:
        raise Exception("Unknown filter type")

    for i in range(len(values)):
        feedforward_taps = np.convolve(feedforward_taps, b[:, i])

    return feedforward_taps, feedback_taps


def _get_coefficients_for_delay(tau, full_taps_x, Ts=1.0):
    full_taps = np.sinc((full_taps_x - tau) / Ts)
    full_taps = _round_taps_close_to_zero(full_taps)
    return full_taps


def _round_taps_close_to_zero(taps, accuracy=1e-6):
    taps[np.abs(taps) < accuracy] = 0
    return taps


def _check_hardware_limitation(qop_version: QOPVersion, feedforward_taps: List, feedback_taps: List):
    def _warning_on_one_line(message, category, filename, lineno, file=None, line=None):
        return "%s:%s: %s: %s\n" % (filename, lineno, category.__name__, message)

    warnings.formatwarning = _warning_on_one_line
    warnings.simplefilter("always", UserWarning)

    feedback_taps = np.array(feedback_taps)
    feedforward_taps = np.array(feedforward_taps)

    # Check limitation on the number of feedforward taps
    max_feedforward_len = qop_version.value["feedforward_length"](len(feedback_taps))
    if len(feedforward_taps) > max_feedforward_len:
        extra_taps = feedforward_taps[max_feedforward_len:]
        final_taps = feedforward_taps[:max_feedforward_len]
        if np.max(np.sum(extra_taps)) / np.max(np.sum(final_taps)) > 0.10:  # Contribution is more than 2%
            raise RuntimeError(
                f"The number of feedforward taps exceed the maximum length of "
                f"{qop_version.value['feedforward_length'](len(feedback_taps))}.\nRemoved all taps afterwards."
                f" The contribution from the removed taps was "
                f"{np.max(np.sum(extra_taps)) / np.max(np.sum(final_taps)) * 100:.1f}%."
            )
        elif np.max(np.sum(extra_taps)) / np.max(np.sum(final_taps)) > 0.02:  # Contribution is more than 2%
            warnings.warn(
                f"The number of feedforward taps exceed the maximum length of "
                f"{qop_version.value['feedforward_length'](len(feedback_taps))}.\nRemoved all taps afterwards."
                f" The contribution from the removed taps was "
                f"{np.max(np.sum(extra_taps)) / np.max(np.sum(final_taps)) * 100:.1f}%."
            )
        feedforward_taps = final_taps

    # Check limitation on the max value of a feedback tap
    if np.any(np.abs(feedback_taps) > qop_version.value["feedback_max"]):
        feedback_taps[feedback_taps > qop_version.value["feedback_max"]] = qop_version.value["feedback_max"]
        feedback_taps[feedback_taps < -qop_version.value["feedback_max"]] = -qop_version.value["feedback_max"]
        warn(
            f"The feedback taps reached the maximum value of {qop_version.value['feedback_max']} which may result in "
            f"a non-optimal filter implementation."
        )

    # Check limitation on the max value of a feedforward tap
    max_value = max(np.abs(feedforward_taps))
    if max_value > qop_version.value["feedforward_max"]:
        feedforward_taps = qop_version.value["feedforward_max"] * feedforward_taps / max_value
        warn(
            f"The feedforward taps reached the maximum value of {qop_version.value['feedforward_max']}.\n"
            f"The coefficients are scaled down to stay within the valid range which reduces the outputted amplitude of"
            f" the pulses played through the filtered port by a factor of "
            f"{max_value / qop_version.value['feedforward_max']:.3f}."
        )

    return list(feedforward_taps), list(feedback_taps)



import numpy as np
from qm import ControllerConnection, InterOpxChannel


def create_simulator_controller_connections(
    n_controllers, actual_controllers=None, n_connections=12, print_debug=False
):
    """
    Creates a list of :class:`~qm.simulate.interface.ControllerConnection` objects with the Inter-controllers
    connections between controllers in a way which maximizes the optical links connectivity.
    For use inside the :class:`~qm.simulate.interface.SimulationConfig` for simulating the controllers` behavior.

    :param n_controllers: The number of controllers
    :type n_controllers: int
    :param actual_controllers: A list of the controllers actually being used in the simulation. For `con1` & `con2` use [1,2].
    :type actual_controllers: List[int]
    :param n_connections: The number of optical connectors each controller has, default is 12
    :type n_connections: int
    :param print_debug: If true, prints the connections to the terminal
    :type print_debug: bool
    :returns: The ControllerConnection object

    """
    if n_controllers == 1:
        return []
    if not actual_controllers:
        actual_controllers = [i + 1 for i in range(n_controllers)]
    controller_connections = []
    unused_connection = np.ones((n_controllers, n_connections), dtype=bool)
    while unused_connection.any():
        for i in range(n_controllers):
            for j in range(i + 1, n_controllers):
                first_con_port = np.nonzero(unused_connection[i, :])[0]
                if first_con_port.size == 0:
                    break
                first_con_port = first_con_port[0].item()
                second_con_port = np.nonzero(unused_connection[j, :])[0]
                if second_con_port.size == 0:
                    break
                second_con_port = second_con_port[0].item()
                if i + 1 in actual_controllers and j + 1 in actual_controllers:
                    controller_connections.append(
                        ControllerConnection(
                            InterOpxChannel(f"con{i + 1}", first_con_port),
                            InterOpxChannel(f"con{j + 1}", second_con_port),
                        )
                    )
                    if print_debug:
                        print(f"con{i + 1}:{first_con_port} <-> con{j + 1}, {second_con_port}")
                elif print_debug:
                    print(f"con{i + 1}:{first_con_port} <-> con{j + 1}, {second_con_port} - Skipped")
                unused_connection[i, first_con_port] = False
                unused_connection[j, second_con_port] = False

    return controller_connections



from ._cirq import import_cirq
from .simple_tableau import SimpleTableau

import dataclasses
import os
import pathlib
import pickle
import random
from typing import Set, List

import numpy as np

cirq = import_cirq()


q1, q2 = cirq.LineQubit.range(1, 3)

C1_reduced = [
    cirq.PhasedXZGate(axis_phase_exponent=0, x_exponent=0, z_exponent=0),
    cirq.PhasedXZGate(axis_phase_exponent=0, x_exponent=-0.5, z_exponent=0),
    cirq.PhasedXZGate(axis_phase_exponent=0.5, x_exponent=-0.5, z_exponent=1),
    cirq.PhasedXZGate(axis_phase_exponent=0.5, x_exponent=-0.5, z_exponent=-0.5),
    cirq.PhasedXZGate(axis_phase_exponent=0, x_exponent=0.5, z_exponent=0.5),
    cirq.PhasedXZGate(axis_phase_exponent=0, x_exponent=0, z_exponent=0.5),
]

S1 = [
    cirq.PhasedXZGate(axis_phase_exponent=0, x_exponent=0, z_exponent=0),
    cirq.PhasedXZGate(axis_phase_exponent=0, x_exponent=0.5, z_exponent=0.5),
    cirq.PhasedXZGate(axis_phase_exponent=0.5, x_exponent=-0.5, z_exponent=-0.5),
]

pauli = [
    cirq.PhasedXZGate(axis_phase_exponent=0, x_exponent=0, z_exponent=0),  # I
    cirq.PhasedXZGate(axis_phase_exponent=0, x_exponent=1.0, z_exponent=0),  # X
    cirq.PhasedXZGate(axis_phase_exponent=0.5, x_exponent=1.0, z_exponent=0),  # Y
    cirq.PhasedXZGate(axis_phase_exponent=0, x_exponent=0, z_exponent=1.0),  # Z
]

pauli_phase = [
    [0, 0],  # I
    [0, 1],  # X
    [1, 1],  # Y
    [1, 0],  # Z
]

native_2_qubit_gates = {
    "sqr_iSWAP": {
        "CNOT": [
            cirq.PhasedXZGate(axis_phase_exponent=0.5, x_exponent=0.5, z_exponent=-1)(q1),
            cirq.PhasedXZGate(axis_phase_exponent=0, x_exponent=0, z_exponent=1)(q2),
            cirq.ISWAP(q1, q2) ** 0.5,
            cirq.PhasedXZGate(axis_phase_exponent=0, x_exponent=1, z_exponent=0)(q1),
            cirq.PhasedXZGate(axis_phase_exponent=0, x_exponent=0, z_exponent=0)(q2),
            cirq.ISWAP(q1, q2) ** 0.5,
            cirq.PhasedXZGate(axis_phase_exponent=0.5, x_exponent=0.5, z_exponent=0.5)(q1),
            cirq.PhasedXZGate(axis_phase_exponent=-1, x_exponent=0.5, z_exponent=1)(q2),
        ],  # compilation of CNOT in terms of phased XZ and sqiSWAP
        "iSWAP": [
            cirq.ISWAP(q1, q2) ** 0.5,
            cirq.ISWAP(q1, q2) ** 0.5,
        ],  # compilation of iSWAP in terms of phased XZ and sqiSWAP
        "SWAP": [
            cirq.PhasedXZGate(axis_phase_exponent=0, x_exponent=0.5, z_exponent=0.5)(q1),
            cirq.PhasedXZGate(axis_phase_exponent=0.5, x_exponent=0.5, z_exponent=0)(q2),
            cirq.ISWAP(q1, q2) ** 0.5,
            cirq.PhasedXZGate(axis_phase_exponent=-1, x_exponent=0.5, z_exponent=1)(q1),
            cirq.PhasedXZGate(axis_phase_exponent=-1, x_exponent=0.5, z_exponent=1)(q2),
            cirq.ISWAP(q1, q2) ** 0.5,
            cirq.PhasedXZGate(axis_phase_exponent=0.5, x_exponent=0.5, z_exponent=0)(q1),
            cirq.PhasedXZGate(axis_phase_exponent=0.5, x_exponent=0.5, z_exponent=0)(q2),
            cirq.ISWAP(q1, q2) ** 0.5,
            cirq.PhasedXZGate(axis_phase_exponent=0, x_exponent=0, z_exponent=-0.5)(q1),
            cirq.PhasedXZGate(axis_phase_exponent=0, x_exponent=0, z_exponent=1)(q2),
        ],  # compilation of SWAP in terms of phased XZ and sqiSWAP
    },
    # TODO: add more implementations
    "CNOT": {
        "CNOT": [cirq.CNOT(q1, q2)],
        "iSWAP": [
            cirq.PhasedXZGate(axis_phase_exponent=-1.0, x_exponent=0.5, z_exponent=-0.5)(q1),  # S + H -> PhasedXZ
            cirq.PhasedXZGate(axis_phase_exponent=0.0, x_exponent=0.0, z_exponent=0.5)(q2),  # S
            cirq.CNOT(q1, q2),
            cirq.CNOT(q2, q1),
            cirq.PhasedXZGate(axis_phase_exponent=0.0, x_exponent=0.0, z_exponent=0.0)(q1),  # I
            cirq.PhasedXZGate(axis_phase_exponent=-0.5, x_exponent=0.5, z_exponent=-1.0)(q2),  # H -> PhasedXZ
        ],
        "SWAP": [cirq.CNOT(q2, q1), cirq.CNOT(q1, q2), cirq.CNOT(q2, q1)],
    },
    "CZ": {
        "CNOT": [
            cirq.PhasedXZGate(axis_phase_exponent=-0.5, x_exponent=0.5, z_exponent=-1.0)(q2),
            cirq.PhasedXZGate(axis_phase_exponent=0.0, x_exponent=0.0, z_exponent=0.0)(q1),
            cirq.CZ(q1, q2),
            cirq.PhasedXZGate(axis_phase_exponent=-0.5, x_exponent=0.5, z_exponent=-1.0)(q2),
            cirq.PhasedXZGate(axis_phase_exponent=0.0, x_exponent=0.0, z_exponent=0.0)(q1),
        ],
        "iSWAP": [
            cirq.PhasedXZGate(axis_phase_exponent=-1.0, x_exponent=0.5, z_exponent=-0.5)(q1),
            cirq.PhasedXZGate(axis_phase_exponent=-1.0, x_exponent=0.5, z_exponent=-0.5)(q2),
            cirq.CZ(q1, q2),
            cirq.PhasedXZGate(axis_phase_exponent=-0.5, x_exponent=0.5, z_exponent=-1.0)(q1),
            cirq.PhasedXZGate(axis_phase_exponent=-0.5, x_exponent=0.5, z_exponent=-1.0)(q2),
            cirq.CZ(q1, q2),
            cirq.PhasedXZGate(axis_phase_exponent=-0.5, x_exponent=0.5, z_exponent=-1.0)(q1),
            cirq.PhasedXZGate(axis_phase_exponent=-0.5, x_exponent=0.5, z_exponent=-1.0)(q2),
        ],
        "SWAP": [
            cirq.PhasedXZGate(axis_phase_exponent=-0.5, x_exponent=0.5, z_exponent=-1.0)(q2),
            cirq.PhasedXZGate(axis_phase_exponent=0.0, x_exponent=0.0, z_exponent=0.0)(q1),
            cirq.CZ(q1, q2),
            cirq.PhasedXZGate(axis_phase_exponent=-0.5, x_exponent=0.5, z_exponent=-1.0)(q1),
            cirq.PhasedXZGate(axis_phase_exponent=-0.5, x_exponent=0.5, z_exponent=-1.0)(q2),
            cirq.CZ(q1, q2),
            cirq.PhasedXZGate(axis_phase_exponent=-0.5, x_exponent=0.5, z_exponent=-1.0)(q1),
            cirq.PhasedXZGate(axis_phase_exponent=-0.5, x_exponent=0.5, z_exponent=-1.0)(q2),
            cirq.CZ(q1, q2),
            cirq.PhasedXZGate(axis_phase_exponent=-0.5, x_exponent=0.5, z_exponent=-1.0)(q2),
            cirq.PhasedXZGate(axis_phase_exponent=0.0, x_exponent=0.0, z_exponent=0.0)(q1),
        ],
    },
}


# Conversion from unitary / cirq to tableau
I = np.array([[1, 0], [0, 1]])  # identity
X = np.array([[0, 1], [1, 0]])  # pi x
Y = np.array([[0, -1j], [1j, 0]])  # pi y
Z = np.array([[1, 0], [0, -1]])  # pi z

paulis = [I, X, Y, Z]
twoQBPaulis = [np.kron(i, j) for i in paulis for j in paulis]

symplecticTable = [
    [0, 0, 0, 0],  # II
    [0, 0, 1, 0],  # IX
    [0, 0, 1, 1],  # IY
    [0, 0, 0, 1],  # IZ
    [1, 0, 0, 0],  # XI
    [1, 0, 1, 0],  # XX
    [1, 0, 1, 1],  # XY
    [1, 0, 0, 1],  # XZ
    [1, 1, 0, 0],  # YI
    [1, 1, 1, 0],  # YX
    [1, 1, 1, 1],  # YY
    [1, 1, 0, 1],  # YZ
    [0, 1, 0, 0],  # ZI
    [0, 1, 1, 0],  # ZX
    [0, 1, 1, 1],  # ZY
    [0, 1, 0, 1],
]  # ZZ


def get_pauli_prod(m):
    # input: tensor product of two Pauli matrices
    # output: tableau column of m

    for i, p in enumerate(twoQBPaulis):
        prod = m @ p
        if np.trace(prod) > 3.9:
            return symplecticTable[i], 0
        if np.trace(prod) < -3.9:
            return symplecticTable[i], 1


def tableau_from_unitary(m):
    # Turns a two-qubit unitary into the full tableau representation
    # input: two-qubit unitary m
    # outputs: simple tableau
    s = np.zeros([4, 4])
    p = np.zeros(4)

    s[:, 0], p[0] = get_pauli_prod(m @ np.kron(X, I) @ m.H)
    s[:, 1], p[1] = get_pauli_prod(m @ np.kron(Z, I) @ m.H)
    s[:, 2], p[2] = get_pauli_prod(m @ np.kron(I, X) @ m.H)
    s[:, 3], p[3] = get_pauli_prod(m @ np.kron(I, Z) @ m.H)

    return SimpleTableau(s, p)


def tableau_from_cirq(gates: List[cirq.GateOperation]) -> SimpleTableau:
    return tableau_from_unitary(np.matrix(cirq.Circuit(gates).unitary()))


#########################################################


def combine_to_phased_x_z(first_gate: cirq.GateOperation, second_gate: cirq.GateOperation) -> cirq.GateOperation:
    unitary = cirq.Circuit([first_gate, second_gate]).unitary()
    if unitary.shape != (2, 2):
        raise RuntimeError("Cannot combine multi qubit gate to PhasedXZ")
    return cirq.PhasedXZGate.from_matrix(unitary)(first_gate.qubits[0])


@dataclasses.dataclass
class GateCommand:
    type: str
    q1: tuple
    q2: tuple

    def get_qubit_ops(self, q):
        if q == 0:
            return self.q1
        elif q == 1:
            return self.q2
        else:
            raise RuntimeError("q should be 0 or 1")


class _GateDatabase:
    def __init__(self):
        self._commands, self._tableaus, self._symplectic_range, self._pauli_range = self._gen_commands_and_tableaus()

    @staticmethod
    def _gen_commands_and_tableaus():
        compilation_path = pathlib.Path(os.path.dirname(os.path.abspath(__file__))) / "symplectic_compilation_XZ.pkl"
        with open(compilation_path, "rb") as f:
            compilation = pickle.load(f)
        symplectics = compilation["symplectics"]
        phases = compilation["phases"]
        commands = compilation["commands"]

        rb_commands = []
        for command in commands:
            if command[0] == "C1's":
                rb_commands.append(GateCommand("C1", (command[1],), (command[2],)))
            elif command[0] == "CNOT's":
                rb_commands.append(GateCommand("CNOT", (command[1], command[3]), (command[2], command[4])))
            elif command[0] == "iSWAP's":
                rb_commands.append(GateCommand("iSWAP", (command[1], command[3]), (command[2], command[4])))
            elif command[0] == "SWAP's":
                rb_commands.append(GateCommand("SWAP", (command[1],), (command[2],)))

        # Generate Paulis:
        for i1 in range(len(pauli)):
            for i2 in range(len(pauli)):
                rb_commands.append(GateCommand("PAULI", (i1,), (i2,)))

        tableaus = []
        for i in range(len(symplectics)):
            tableaus.append(SimpleTableau(symplectics[i], phases[i]))

        # Generate Paulis:
        for i1 in range(len(pauli)):
            for i2 in range(len(pauli)):
                tableaus.append(SimpleTableau(np.eye(4), pauli_phase[i1] + pauli_phase[i2]))

        symplectic_range = (0, len(commands))
        pauli_range = (len(commands), len(rb_commands))
        return rb_commands, tableaus, symplectic_range, pauli_range

    @property
    def commands(self):
        return self._commands

    @property
    def tableaus(self):
        return self._tableaus

    def get_command(self, gate_id) -> GateCommand:
        return self._commands[gate_id]

    def get_tableau(self, gate_id) -> SimpleTableau:
        return self._tableaus[gate_id]

    def rand_symplectic(self):
        return random.randrange(*self._symplectic_range)

    def rand_pauli(self):
        return random.randrange(*self._pauli_range)

    def get_interleaving_gate(self):
        return self._pauli_range[1]

    def find_symplectic_gate_id_by_tableau_g(self, tableau: SimpleTableau):
        tableaus = self._tableaus[self._symplectic_range[0] : self._symplectic_range[1]]
        return next(i for i, x in enumerate(tableaus) if np.array_equal(x.g, tableau.g))

    def find_pauli_gate_id_by_tableau_alpha(self, tableau: SimpleTableau):
        tableaus = self._tableaus[self._pauli_range[0] : self._pauli_range[1]]
        return self._pauli_range[0] + next(i for i, x in enumerate(tableaus) if np.array_equal(x.alpha, tableau.alpha))


gate_db = _GateDatabase()


class GateGenerator:
    two_qubit_imp_priority = {  # TODO: verify this priority table
        "CNOT": ["CNOT", "CZ", "iSWAP", "sqr_iSWAP"],
        "iSWAP": ["iSWAP", "sqr_iSWAP", "CNOT", "CZ"],
        "SWAP": ["CNOT", "CZ", "iSWAP", "sqr_iSWAP"],
    }

    def __init__(self, native_two_qubit_gates: Set[str]):
        self._two_qubit_dict = self._generate_two_qubit_dict(native_two_qubit_gates)

    @staticmethod
    def _generate_two_qubit_dict(native_two_qubit_gates: Set[str]) -> dict:
        two_qubit_dict = {}
        for k, v in GateGenerator.two_qubit_imp_priority.items():
            available_imp = [x for x in v if x in native_two_qubit_gates]
            if len(available_imp) == 0 or available_imp[0] not in native_2_qubit_gates.keys():
                raise RuntimeError(f"Cannot implement gate '{k}' with provided native two qubit gates")
            two_qubit_dict[k] = available_imp[0]
        return two_qubit_dict

    @staticmethod
    def _reduce_gate(gate: List[cirq.GateOperation]):
        qubit_ops = {1: None, 2: None}
        output = []

        def append_qubit_ops():
            for q in [1, 2]:
                if qubit_ops[q] is not None:
                    output.append(qubit_ops[q])
                qubit_ops[q] = None

        for op in gate:
            if len(op.qubits) == 1:
                prev_op = qubit_ops[op.qubits[0].x]
                qubit_ops[op.qubits[0].x] = op if prev_op is None else combine_to_phased_x_z(prev_op, op)
            else:
                append_qubit_ops()
                output.append(op)
        append_qubit_ops()
        return output

    def generate(self, cmd_id):
        gate = []
        command = gate_db.get_command(cmd_id)
        two_qubit_imp = self._two_qubit_dict[command.type] if command.type in self._two_qubit_dict else None
        if command.type == "C1":
            gate.append(C1_reduced[command.q1[0]](q1))
            gate.append(C1_reduced[command.q2[0]](q2))
        elif command.type == "CNOT":
            gate.append(C1_reduced[command.q1[0]](q1))
            gate.append(C1_reduced[command.q2[0]](q2))
            gate.extend(native_2_qubit_gates[two_qubit_imp]["CNOT"])
            gate.append(S1[command.q1[1]](q1))
            gate.append(S1[command.q2[1]](q2))
        elif command.type == "iSWAP":
            gate.append(C1_reduced[command.q1[0]](q1))
            gate.append(C1_reduced[command.q2[0]](q2))
            gate.extend(native_2_qubit_gates[two_qubit_imp]["iSWAP"])
            gate.append(S1[command.q1[1]](q1))
            gate.append(S1[command.q2[1]](q2))
        elif command.type == "SWAP":
            gate.append(C1_reduced[command.q1[0]](q1))
            gate.append(C1_reduced[command.q2[0]](q2))
            gate.extend(native_2_qubit_gates[two_qubit_imp]["SWAP"])
        elif command.type == "PAULI":
            gate.append(pauli[command.q1[0]](q1))
            gate.append(pauli[command.q2[0]](q2))
        else:
            raise RuntimeError(f"unknown command {command.type}")
        return self._reduce_gate(gate)
